from __future__ import annotations

from os import PathLike
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Literal,
    Mapping,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

import pyarrow as pa

from chalk.features import DataFrame, Feature, FeatureConverter, Features
from chalk.sql._internal.incremental import IncrementalSettings
from chalk.utils.duration import Duration

if TYPE_CHECKING:
    import sqlalchemy
    from sqlalchemy.engine import URL
    from sqlalchemy.orm import Session

    from chalk.sql.finalized_query import FinalizedChalkQuery

TTableIngestProtocol = TypeVar("TTableIngestProtocol", bound="TableIngestProtocol")


class StringChalkQueryProtocol(Protocol):
    def execute(self) -> DataFrame:
        """Materialize the query.

        Chalk queries are lazy, which allows Chalk
        to perform performance optimizations like push-down filters.
        Instead of calling execute, consider returning this query from
        a resolver as an intermediate feature, and processing that
        intermediate feature in a different resolver.

        Returns
        -------
        DataFrame
            A `DataFrame` with the results of the query.
        """
        ans = self.all().execute()
        assert isinstance(ans, DataFrame)
        return ans

    def one_or_none(self) -> FinalizedChalkQuery:
        """Return at most one result or raise an exception.

        Returns `None` if the query selects no rows. Raises if
        multiple object identities are returned, or if multiple
        rows are returned for a query that returns only scalar
        values as opposed to full identity-mapped entities.

        Returns
        -------
        FinalizedChalkQuery
            A query that can be returned from a resolver.
        """
        ...

    def one(self) -> FinalizedChalkQuery:
        """Return exactly one result or raise an exception.

        Returns
        -------
        FinalizedChalkQuery
            A query that can be returned from a resolver.
        """
        ...

    def all(self) -> FinalizedChalkQuery:
        """Return the results represented by this Query as a list.

        Returns
        -------
        FinalizedChalkQuery
            A query that can be returned from a resolver.
        """
        ...

    @overload
    def incremental(
        self,
        *,
        incremental_column: str,
        lookback_period: Duration = "0s",
        mode: Literal["row", "group"] = "row",
    ) -> FinalizedChalkQuery:
        ...

    @overload
    def incremental(
        self,
        *,
        incremental_column: None = None,
        lookback_period: None = None,
        mode: Literal["parameter"],
    ) -> FinalizedChalkQuery:
        ...

    def incremental(
        self,
        *,
        incremental_column: Optional[str] = None,
        lookback_period: Optional[Duration] = "0s",
        mode: Literal["row", "group", "parameter"] = "row",
    ) -> FinalizedChalkQuery:
        """Operates like `.all()`, but tracks `previous_latest_row_timestamp` between query executions in
        order to limit the amount of data returned.

        `previous_latest_row_timestamp` will be set the start of the query execution, or if you return
        a `FeatureTime`-mapped column, Chalk will update `previous_latest_row_timestamp` to the maximum
        observed `FeatureTime` value.

        In `"row"` mode:
        `incremental_column` MUST be set.

        Returns the results represented by this query as a list (like `.all()`), but modifies the query to
        only return "new" results, by adding a clause that looks like:

            ```
            WHERE <incremental_column> >= <previous_latest_row_timestamp> - <lookback_period>
            ```

        In `"group"` mode:
        `incremental_column` MUST be set.

        Returns the results represented by this query as a list (like `.all()`), but modifies the query to
        only results from "groups" which have changed since the last run of the query.

        This works by (1) parsing your query, (2) finding the "group keys", (3) selecting only changed groups.
        Concretely:

            ```
            SELECT user_id, sum(amount) as sum_amount
            FROM payments
            GROUP BY user_id
            ```

        would be rewritten like this:

            ```
            SELECT user_id, sum(amount) as sum_amount
            FROM payments
            WHERE user_id in (
                SELECT DISTINCT(user_id)
                FROM payments WHERE created_at >= <previous_latest_row_timestamp> - <lookback_period>
            )
            GROUP BY user_id
            ```

        In `"parameter"` mode:
        `incremental_column` WILL BE IGNORED.

        This mode is for cases where you want full control of incrementalization. Chalk will not manipulate your query.
        Chalk will include a query parameter named `"chalk_incremental_timestamp"`. Depending on your SQL
        dialect, you can use this value to incrementalize your query with `:chalk_incremental_timestamp` or
        `%(chalk_incremental_timestamp)s`.

        Parameters
        ----------
        incremental_column
            This should reference a timestamp column in your underlying table, typically something
            like `"updated_at"`, `"created_at"`, `"event_time"`, etc.
        lookback_period
            Defaults to `0`, which means we only return rows that are strictly newer than
            the last observed row.
        mode
            Defaults to `"row"`, which indicates that only rows newer than the last observed row should be
            considered. When set to `"group"`, Chalk will only ingest features from groups which are newer
            than the last observation time. This requires that the query is grouped by a primary key.

        Returns
        -------
        FinalizedChalkQuery
            A query that can be returned from a resolver.
        """
        ...


class ChalkQueryProtocol(Protocol):
    def first(self) -> FinalizedChalkQuery:
        """Return the first result of this Query or None if the result doesn't contain any row.

        Returns
        -------
        FinalizedChalkQuery
            A query that can be returned from a resolver.
        """
        ...

    def one_or_none(self) -> FinalizedChalkQuery:
        """Return at most one result or raise an exception.

        Returns `None` if the query selects no rows. Raises if
        multiple object identities are returned, or if multiple
        rows are returned for a query that returns only scalar
        values as opposed to full identity-mapped entities.

        Returns
        -------
        FinalizedChalkQuery
            A query that can be returned from a resolver.
        """
        ...

    def one(self) -> FinalizedChalkQuery:
        """Return exactly one result or raise an exception.

        Returns
        -------
        FinalizedChalkQuery
            A query that can be returned from a resolver.
        """
        ...

    def all(self) -> FinalizedChalkQuery:
        """Return the results represented by this Query as a `DataFrame`.

        Returns
        -------
        FinalizedChalkQuery
            A query that can be returned from a resolver.
        """
        ...

    def incremental(
        self,
        lookback_period: Duration = "0s",
        mode: str = "row",
        incremental_column: Optional[Union[str, Feature]] = None,
    ) -> FinalizedChalkQuery:
        """Operates like `.all()`, but tracks `previous_latest_row_timestamp` between query executions in
        order to limit the amount of data returned.

        `previous_latest_row_timestamp` will be set the start of the query execution, or if you return
        a `FeatureTime`-mapped column, Chalk will update `previous_latest_row_timestamp` to the maximum
        observed `FeatureTime` value.

        In `"row"` mode:
        `incremental_column` MUST be set.

        Returns the results represented by this query as a list (like `.all()`), but modifies the query to
        only return "new" results, by adding a clause that looks like:

            ```
            WHERE <incremental_column> >= <previous_latest_row_timestamp> - <lookback_period>
            ```

        In `"group"` mode:
        `incremental_column` MUST be set.

        Returns the results represented by this query as a list (like `.all()`), but modifies the query to
        only results from "groups" which have changed since the last run of the query.

        This works by (1) parsing your query, (2) finding the "group keys", (3) selecting only changed groups.
        Concretely:

            ```
            SELECT user_id, sum(amount) as sum_amount
            FROM payments
            GROUP BY user_id
            ```

        would be rewritten like this:

            ```
            SELECT user_id, sum(amount) as sum_amount
            FROM payments
            WHERE user_id in (
                SELECT DISTINCT(user_id)
                FROM payments WHERE created_at >= <previous_latest_row_timestamp> - <lookback_period>
            )
            GROUP BY user_id
            ```

        In `"parameter"` mode:
        `incremental_column` WILL BE IGNORED.

        This mode is for cases where you want full control of incrementalization. Chalk will not manipulate your query.
        Chalk will include a query parameter named `"chalk_incremental_timestamp"`. Depending on your SQL
        dialect, you can use this value to incrementalize your query with `:chalk_incremental_timestamp` or
        `%(chalk_incremental_timestamp)s`.

        Parameters
        ----------
        incremental_column
            This should reference a timestamp column in your underlying table, typically something
            like `"updated_at"`, `"created_at"`, `"event_time"`, etc.
        lookback_period
            Defaults to `0`, which means we only return rows that are strictly newer than
            the last observed row.
        mode
            Defaults to `"row"`, which indicates that only rows newer than the last observed row should be
            considered. When set to `"group"`, Chalk will only ingest features from groups which are newer
            than the last observation time. This requires that the query is grouped by a primary key.

        Returns
        -------
        FinalizedChalkQuery
            A query that can be returned from a resolver.
        """
        ...

    def filter_by(self, **kwargs: Any) -> "ChalkQueryProtocol":
        """Apply the given filtering criterion to a copy of this Query,
        using keyword expressions.

        Parameters
        ----------
        kwargs
            The column names assigned to the desired values (i.e. `name="Maria"`).

        Returns
        -------
        ChalkQueryProtocol
            A query that can be returned from a resolver or further filtered.

        Examples
        --------
        >>> from chalk.sql import PostgreSQLSource
        >>> session = PostgreSQLSource()
        >>> session.query(UserFeatures(id=UserSQL.id)).filter_by(name="Maria")
        """

        ...

    def filter(self, *criterion: Any) -> "ChalkQueryProtocol":
        """Apply the given filtering criterion to a copy of this Query, using SQL expressions.

        Parameters
        ----------
        criterion
            SQLAlchemy filter criterion

        Returns
        -------
        ChalkQueryProtocol
            A query that can be returned from a resolver or further filtered.
        """
        ...

    def order_by(self, *clauses: Any) -> "ChalkQueryProtocol":
        """Apply one or more ORDER BY criteria to the query and return the newly resulting Query.

        Parameters
        ----------
        clauses
            SQLAlchemy columns.

        Returns
        -------
        ChalkQueryProtocol
            A query that can be returned from a resolver or further filtered.
        """
        ...

    def group_by(self, *clauses: Any) -> "ChalkQueryProtocol":
        ...

    def having(self, criterion: Any) -> "ChalkQueryProtocol":
        ...

    def union(self, *q: "ChalkQueryProtocol") -> "ChalkQueryProtocol":
        ...

    def union_all(self, *q: "ChalkQueryProtocol") -> "ChalkQueryProtocol":
        ...

    def intersect(self, *q: "ChalkQueryProtocol") -> "ChalkQueryProtocol":
        ...

    def intersect_all(self, *q: "ChalkQueryProtocol") -> "ChalkQueryProtocol":
        ...

    def join(self, target: Any, *props: Any, **kwargs: Any) -> "ChalkQueryProtocol":
        ...

    def outerjoin(self, target: Any, *props: Any, **kwargs: Any) -> "ChalkQueryProtocol":
        ...

    def select_from(self, *from_obj: Any) -> "ChalkQueryProtocol":
        ...

    def execute(self):
        """Materialize the query.

        Chalk queries are lazy, which allows Chalk to perform
        performance optimizations like push-down filters.
        Instead of calling execute, consider returning this query from
        a resolver as an intermediate feature, and processing that
        intermediate feature in a different resolver.

        Returns
        -------
        Any
            The raw result of executing the query.
        """
        return self.all().execute()


class BaseSQLSourceProtocol(Protocol):
    def query_string(
        self,
        query: str,
        fields: Optional[Mapping[str, Union[Feature, str, Any]]] = None,
        args: Optional[Mapping[str, str]] = None,
    ) -> StringChalkQueryProtocol:
        """Run a query from a SQL string.

        Parameters
        ----------
        query
            The query that you'd like to run.
        fields
            A mapping from the column names selected to features.
        args
            Any args in the sql string specified by `query` need
            to have corresponding value assignments in `args`.

        Returns
        -------
        StringChalkQueryProtocol
            A query that can be returned from a `@online` or `@offline` resolver.
        """
        ...

    def query_sql_file(
        self,
        path: Union[str, bytes, PathLike],
        fields: Optional[Mapping[str, Union[Feature, str, Any]]] = None,
        args: Optional[Mapping[str, str]] = None,
    ) -> StringChalkQueryProtocol:
        """Run a query from a SQL file.

        This method allows you to query the SQL file within a Python
        resolver. However, Chalk can also infer resolvers from SQL files.
        See https://docs.chalk.ai/docs/sql#sql-file-resolvers
        for more information.

        Parameters
        ----------
        path
            The path to the file with the sql file,
            relative to the caller's file, or to the
            directory that your `chalk.yaml` file lives in.
        fields
            A mapping from the column names selected to features.
        args
            Any args in the sql file specified by `path` need
            to have corresponding value assignments in `args`.

        Returns
        -------
        StringChalkQueryProtocol
            A query that can be returned from a `@online` or `@offline` resolver.
        """
        ...

    def query(self, *entities: Any) -> ChalkQueryProtocol:
        """Query using a `SQLAlchemy` model.

        Parameters
        ----------
        entities
            Arguments as would normally be passed to a `SQLAlchemy`.

        Returns
        -------
        ChalkQueryProtocol
            A query that can be returned from a resolver.
        """
        ...

    def local_engine_url(self) -> URL:
        ...

    def engine_args(self) -> Mapping[str, Any]:
        return {}

    def get_engine(self) -> sqlalchemy.engine.Engine:
        """Get an SQLAlchemy Engine. The engine will be created and cached on the first call of this method.

        Returns
        -------
        sqlalchemy.engine.Engine
            A SQLAlchemy engine.
        """
        ...

    def warm_up(self) -> None:
        """Warm up the Engine by issuing a simple query."""
        ...

    def raw_session(self) -> Session:
        ...

    def execute_query_to_pyarrow_tables(
        self,
        finalized_query: FinalizedChalkQuery,
        columns_to_converters: Callable[[List[str]], Dict[str, FeatureConverter]],
        connection: Optional[Any] = None,
    ) -> Iterator[pa.Table]:
        ...

    def execute_query_to_batches(
        self,
        finalized_query: FinalizedChalkQuery,
        columns_to_converters: Callable[[List[str]], Dict[str, FeatureConverter]],
        connection: Optional[Any] = None,
    ) -> List[pa.Table]:
        ...

    def execute_query(
        self,
        finalized_query: FinalizedChalkQuery,
        columns_to_converters: Callable[[List[str]], Dict[str, FeatureConverter]],
        connection: Optional[Any] = None,
    ) -> pa.Table:
        """Execute a query to a `pa.Table`.

        Parameters
        ----------
        finalized_query
            The query to execute
        columns_to_converters
            A function that, given the list of column names returned from the database,
            produces a mapping of columns to converters. Columns not in the resulting dictionary should be ignored.
        connection
            A connection to execute the query under. If not specified, then a new connection is acquired.

        Returns
        -------
        pa.Table
            A `pa.Table` containing the results.
        """
        ...

    def compile_query(self, finalized_query: FinalizedChalkQuery) -> Tuple[str, Sequence[Any], Mapping[str, Any]]:
        """Compile a query into a string and the (positional parameters, named parameters)

        Returns
        -------
        tuple[str, Sequence[Any], dict[str, Any]]
            The compiled SQL query.
        """
        ...


class TableIngestProtocol(Protocol):
    def with_table(
        self: Type[TTableIngestProtocol],
        *,
        name: str,
        features: Type[Union[Features, Any]],
        ignore_columns: Optional[List[str]] = None,
        ignore_features: Optional[List[Union[str, Any]]] = None,
        require_columns: Optional[List[str]] = None,
        require_features: Optional[List[Union[str, Any]]] = None,
        column_to_feature: Optional[Dict[str, Any]] = None,
        cdc: Optional[Union[bool, IncrementalSettings]] = None,
    ) -> TTableIngestProtocol:
        """Automatically ingest a table.

        Parameters
        ----------
        name
            The name of the table to ingest.
        features
            The feature class that this table should be mapping to, e.g. `User`.
        ignore_columns
            Columns in the table that should be ignored, and not mapped to features,
            even if there is a matching name.
        ignore_features
            Features on the feature class that should be ignored, and not mapped to columns,
            even if there is a matching name.
        require_columns
            Columns that must exist in the mapping.
        require_features
            Features that must exist in the mapping.
        column_to_feature
            Explicit mapping of columns to features for names that do not match.
        cdc
            Settings for incrementally ingesting the table.

        Examples
        --------
        >>> from chalk.sql import PostgreSQLSource
        >>> from chalk.features import features
        >>> PostgreSQLSource().with_table(
        ...     name="users",
        ...     features=User,
        ... ).with_table(
        ...     name="accounts",
        ...     features=Account,
        ...     # Override one of the column mappings.
        ...     column_to_feature={
        ...         "acct_id": Account.id,
        ...     },
        ... )
        """
        ...


class SQLSourceWithTableIngestProtocol(TableIngestProtocol, BaseSQLSourceProtocol, Protocol):
    ...


__all__ = [
    "StringChalkQueryProtocol",
    "ChalkQueryProtocol",
    "BaseSQLSourceProtocol",
    "TableIngestProtocol",
    "SQLSourceWithTableIngestProtocol",
]



from typing import Any, Type, TypeVar, cast

from typing_extensions import Annotated

from chalk.features.feature_wrapper import unwrap_feature

T = TypeVar("T")


class PrimaryMeta(type):
    def __getitem__(self, item: Type[T]) -> Type[T]:
        return cast(Type[T], Annotated[item, "__chalk_primary__"])


Primary = PrimaryMeta("Primary", (object,), {})
"""Marks a feature as the primary feature for a feature class.

Features named `id` on feature classes without an explicit primary
feature are declared primary keys by default, and don't need to be
marked with `Primary`.

If you have primary key feature with a name other than `id`,
you can use this marker to indicate the primary key.

Examples
--------
>>> @features
... class User:
...     uid: Primary[int]
"""


def is_primary(f: Any) -> bool:
    """Determine whether a feature is a primary key.

    Parameters
    ----------
    f
        A feature (i.e. `User.email`)

    Raises
    ------
    TypeError
        If `f` is not a feature.

    Returns
    -------
    bool
        `True` if `f` is primary and `False` otherwise.

    Examples
    --------
    >>> from chalk.features import features
    >>> @features
    ... class User:
    ...     uid: Primary[int]
    ...     email: str
    >>> assert is_primary(User.uid)
    >>> assert not is_primary(User.email)
    """
    return unwrap_feature(f).primary



import collections
from collections import defaultdict
from typing import Dict, List, Mapping, NewType

from chalk.parsed.duplicate_input_gql import (
    GraphLogSeverity,
    ProjectSettingsGQL,
    UpdateGraphError,
    UpsertFeatureGQL,
    UpsertGraphGQL,
    UpsertResolverGQL,
)
from chalk.utils.duration import parse_chalk_duration
from chalk.utils.string import oxford_comma_list, to_camel_case


class _LogBuilder:
    def __init__(self):
        self._logs: List[UpdateGraphError] = []

    def add_log(self, header: str, subheader: str, severity: GraphLogSeverity):
        self._logs.append(UpdateGraphError(header=header, subheader=subheader, severity=severity))

    def add_error(self, header: str, subheader: str):
        self.add_log(header=header, subheader=subheader, severity=GraphLogSeverity.ERROR)

    def add_warning(self, header: str, subheader: str):
        self.add_log(header=header, subheader=subheader, severity=GraphLogSeverity.WARNING)

    def add_info(self, header: str, subheader: str):
        self.add_log(header=header, subheader=subheader, severity=GraphLogSeverity.INFO)

    def get_logs(self) -> List[UpdateGraphError]:
        return self._logs


def _validate_primary_key(feature: UpsertFeatureGQL, builder: _LogBuilder):
    if (
        feature.scalarKind is not None
        and feature.scalarKind.primary
        and feature.scalarKind.scalarKind
        not in [
            "str",
            "int",
        ]
    ):
        builder.add_error(
            header=f'Invalid primary key type "{feature.id.fqn}"',
            subheader=f"Primary keys must be integers or strings. "
            f'Feature "{feature.id.fqn}" has a type of {feature.scalarKind.scalarKind}',
        )


def _validate_max_staleness(feature: UpsertFeatureGQL, builder: _LogBuilder):
    if feature.maxStaleness is not None and feature.maxStaleness != "infinity":
        try:
            parse_chalk_duration(feature.maxStaleness)
        except:
            builder.add_error(
                header=f'Could not parse max_staleness for feature "{feature.id.fqn}"',
                subheader=f'Received "{feature.maxStaleness}". See https://docs.chalk.ai/docs/duration for valid formats.',
            )


def _validate_etl_to_online(feature: UpsertFeatureGQL, builder: _LogBuilder):
    if feature.etlOfflineToOnline and feature.maxStaleness is None:
        builder.add_error(
            header=f'Missing max staleness for "{feature.id.fqn}"',
            subheader=f'The feature "{feature.id.fqn}" is set to ETL to online, but doesn\'t specify a max staleness. Any ETL to online would be immediately invalidated.',
        )


def _validate_no_feature_times_as_input(
    fqn_to_feature: Mapping[str, UpsertFeatureGQL],
    resolver: UpsertResolverGQL,
    builder: _LogBuilder,
):
    for i in resolver.inputs or []:
        if i.underlying.fqn not in fqn_to_feature:
            builder.add_error(
                header=f'Resolver "{resolver.fqn}" an unknown feature {i.underlying.fqn}',
                subheader=f"""Please check the inputs to this resolver, and make sure they are all features.""",
            )
        elif fqn_to_feature[i.underlying.fqn].featureTimeKind is not None and resolver.kind != "offline":
            builder.add_error(
                header=f"Feature times cannot be accepted as input to {resolver.kind} resolvers.",
                subheader=f'The resolver "{resolver.fqn}" takes as input the\n'
                f'feature "{i.underlying.fqn}", a feature_time() type.\n'
                f"Feature times can be returned from {resolver.kind} resolvers to\n"
                "indicate data from the past, but cannot be accepted\n"
                "as inputs.",
            )


def _validate_feature_names_unique(features: List[UpsertFeatureGQL], builder: _LogBuilder):
    counter: dict[str, int] = defaultdict(lambda: 0)
    for r in features:
        counter[r.id.fqn] += 1

    for fqn, count in counter.items():
        if count > 1:
            builder.add_error(
                header="Duplicate feature names",
                subheader=f'There are {count} features with the same name of "{fqn}". All features require a '
                f"distinct name",
            )


def _validate_resolver_names_unique(resolvers: List[UpsertResolverGQL], builder: _LogBuilder):
    counter: dict[str, int] = defaultdict(lambda: 0)
    for r in resolvers:
        counter[r.fqn.split(".")[-1]] += 1

    for name, count in counter.items():
        if count > 1:
            builder.add_error(
                header="Duplicate resolver names",
                subheader=f'There are {count} resolvers with the same name of "{name}". All resolvers require a '
                f"distinct name",
            )


def _validate_offline_resolver_has_one(resolver: UpsertResolverGQL, builder: _LogBuilder):
    if resolver.kind != "offline":
        return

    for i in resolver.inputs or []:
        if i.path is not None and len(i.path) > 1:
            full_path = [i.path[0].parent.fqn]
            for p in i.path[1:]:
                full_path.append(p.child.namespace)
            full_path.append(i.path[-1].parentToChildAttributeName)
            pathname = ".".join(full_path)
            builder.add_error(
                header=f"Invalid offline resolver input",
                subheader=f"Offline resolvers cannot require nested of more than one level. "
                + f'The resolver "{resolver.fqn}" takes as input the feature "{pathname}"',
            )


def _validate_resolver_input(resolver: UpsertResolverGQL, builder: _LogBuilder):
    namespaces = set()

    for i in resolver.inputs or []:
        if i.path is not None and len(i.path) > 0:
            namespaces.add(i.path[0].parent.namespace)
        else:
            namespaces.add(i.underlying.namespace)
    namespaces = list(namespaces)
    if len(namespaces) > 1:
        builder.add_error(
            header=f'Resolver "{resolver.fqn}" requires features from multiple namespaces.',
            subheader=f"""All inputs to a resolver should be rooted in the same namespace.
The resolver "{resolver.fqn}" takes inputs in the namespaces {oxford_comma_list(namespaces)}.

If you require features from many feature classes, reference them via their relationships, such as:
  User.bank_account.title,
  User.full_name

{resolver.functionDefinition}
""",
        )


def _validate_resolver_output(resolver: UpsertResolverGQL, builder: _LogBuilder):
    output = resolver.output
    features = output.features or []
    dataframes = output.dataframes or []
    if len(features) == 0 and len(dataframes) == 0:
        builder.add_error(
            header=f'Resolver "{resolver.fqn}" does not define outputs.',
            subheader="See https://docs.chalk.ai/docs/resolver-outputs for information about valid resolver outputs.",
        )

    if len(dataframes) > 1:
        builder.add_error(
            header=f'Resolver "{resolver.fqn}" defines multiple DataFrames as output.',
            subheader="See https://docs.chalk.ai/docs/resolver-outputs for information about valid resolver outputs.",
        )

    if len(features) > 0 and len(dataframes) > 0:
        builder.add_error(
            header=f'Resolver "{resolver.fqn}" returns both relationships and scalar features.',
            subheader="See https://docs.chalk.ai/docs/resolver-outputs for information about valid resolver outputs.",
        )

    # Want to validate this but only have feature id instead of feature
    # if len(resolver.inputs) == 0 and len(dataframes) == 1 and len([pkey for pkey in dataframes[0].columns if pkey.primary]) != 1:
    #     builder.add_error(
    #         header=f'Resolver "{resolver.fqn}" must return a primary feature',
    #         subheader="See https://docs.chalk.ai/docs/resolver-outputs for information about valid resolver outputs.",
    #     )

    if len(features) > 0:
        namespaces = list(set(f.namespace for f in features))
        if len(namespaces) > 1:
            builder.add_error(
                header=f'Resolver "{resolver.fqn}" outputs features from multiple namespaces.',
                subheader=f"""All outputs of a resolver should be rooted in the same namespace.
The resolver "{resolver.fqn}" outputs features in the namespaces {oxford_comma_list(namespaces)}.
See https://docs.chalk.ai/docs/resolver-outputs for information about valid resolver outputs.

{resolver.functionDefinition}
""",
            )


# FIXME CHA-66 we should validate that joins are all pkey on pkey joins. no non-primary keys, no self-joins, no constants
def _validate_joins(feature: UpsertFeatureGQL, builder: _LogBuilder):
    pass


def _validate_feature_names(feature: UpsertFeatureGQL, builder: _LogBuilder):
    if feature.id.name.startswith("__") or feature.id.name.startswith("_chalk"):
        builder.add_error(
            header="Feature uses protected name",
            subheader=f'The feature "{feature.id.fqn}" uses the protected name "{feature.id.name}". '
            'Feature names cannot begin with "_chalk" or "__". Please rename this feature.',
        )
    if feature.id.namespace.startswith("__") or feature.id.name.startswith("_chalk"):
        builder.add_error(
            header="Feature class uses protected namespace",
            subheader=f'The feature "{feature.id.fqn}" belongs to the protected namespace "{feature.id.namespace}". '
            f'Feature namespaces cannot begin with "_chalk" or "__". Please rename this feature set.',
        )


# FIXME: This needs to check for cycles. In particular:
#   def resolve_x(id: User.id) -> DataFrame[User]
# is bogus

_BUILT_IN_SUPPORTED_TYPES = {
    "datetime",
    "date",
    "int",
    "float",
    "str",
    "bool",
    "Enum",
    "__dataclass__",
    "__attrs__",
    "__pydantic__",
}


def _validate_feature_metadata(
    features: List[UpsertFeatureGQL],
    builder: _LogBuilder,
    request_config: ProjectSettingsGQL,
):
    MissingMetadata = NewType("MissingMetadata", Dict[GraphLogSeverity, List[str]])

    class WrappedFeature:
        def __init__(self, feature: UpsertFeatureGQL, missing_metadata: MissingMetadata):
            self.feature = feature
            self.missing_metadata = missing_metadata

    def get_missing_metadata(feature: UpsertFeatureGQL) -> MissingMetadata:
        res: MissingMetadata = collections.defaultdict(list)
        for setting in request_config.validation.feature.metadata:
            metadata_name = setting.name
            severity = GraphLogSeverity(setting.missing.upper())
            if hasattr(feature, metadata_name):
                value = getattr(feature, metadata_name)
                is_missing = value is None or (isinstance(value, list) and len(value) == 0)
                if is_missing:
                    res[severity].append(metadata_name)

        return res

    def get_missing_metadata_log(wrapped_features: List[WrappedFeature], severity: GraphLogSeverity):
        if not wrapped_features:
            return None, None

        missing_messages = []
        feature_str = "feature"
        missing_metadata_header_str = "missing metadata"
        max_name_len = max([len(w.feature.id.name) for w in wrapped_features])
        feature_column_width = max(max_name_len, len(feature_str)) + 1
        get_padding = lambda s: feature_column_width - len(s)

        first_feature = wrapped_features[0].feature
        header = f'"{to_camel_case(first_feature.id.namespace)}" features have missing metadata'
        subheader = f"  Filepath: {first_feature.namespacePath}\n\n"

        for wrapper in wrapped_features:
            missing_metadata = wrapper.missing_metadata.get(severity)
            if not missing_metadata:
                continue
            padding_1 = get_padding(wrapper.feature.id.name)
            missing_metadata_str = ", ".join(missing_metadata)
            missing_messages.append(f"      {wrapper.feature.id.name}{' ' * padding_1}: {missing_metadata_str}")

        if not missing_messages:
            return None, None

        padding_2 = get_padding(feature_str)
        subheader += f"      {'-' * len(feature_str)}{' ' * padding_2}  {'-' * len(missing_metadata_header_str)}\n"
        subheader += f"      {feature_str}{' ' * padding_2}  {missing_metadata_header_str}\n"
        subheader += f"      {'-' * len(feature_str)}{' ' * padding_2}  {'-' * len(missing_metadata_header_str)}\n"
        subheader += "\n".join(missing_messages)

        return header, subheader

    def build_log_for_each_severity(wrapped_features: List[WrappedFeature]):
        if not wrapped_features:
            return

        for severity in GraphLogSeverity:
            header, subheader = get_missing_metadata_log(
                wrapped_features=wrapped_features, severity=GraphLogSeverity(severity)
            )
            if header is None:
                continue
            builder.add_log(header=header, subheader=subheader, severity=GraphLogSeverity(severity))

    def get_wrapped_features(
        namespace_features: List[UpsertFeatureGQL],
    ) -> List[WrappedFeature]:
        wrapped_features = []
        for nf in namespace_features:
            missing_metadata = get_missing_metadata(nf)
            if not missing_metadata:
                continue
            wrapped_features.append(WrappedFeature(feature=nf, missing_metadata=missing_metadata))

        return wrapped_features

    if not (
        request_config
        and request_config.validation
        and request_config.validation.feature
        and request_config.validation.feature.metadata
    ):
        return

    namespace_to_features = collections.defaultdict(list)
    for f in features:
        namespace_to_features[f.id.namespace].append(f)

    for namespace, ns_features in namespace_to_features.items():
        wf = get_wrapped_features(namespace_features=ns_features)
        build_log_for_each_severity(wrapped_features=wf)


def _validate_feature_metadata_config(
    features: List[UpsertFeatureGQL],
    builder: _LogBuilder,
    request_config: ProjectSettingsGQL,
):
    if not (
        request_config
        and request_config.validation
        and request_config.validation.feature
        and request_config.validation.feature.metadata
    ):
        return

    severities_lower = [e.lower() for e in GraphLogSeverity]
    first_feature = features and features[0]

    for missing_metadata in request_config.validation.feature.metadata:
        metadata_name = missing_metadata.name
        severity = missing_metadata.missing

        severity_upper = severity.upper()
        try:
            GraphLogSeverity(severity_upper)
        except ValueError:
            severity_choices = '" or "'.join(severities_lower)
            builder.add_warning(
                header=f'Found invalid log severity "{severity}" config for missing metadata',
                subheader=(
                    f'The required feature metadata "{metadata_name}" is associated with an invalid log severity "{severity}".'
                    f' Please use "{severity_choices}" in chalk.yml'
                ),
            )

        if first_feature and not hasattr(first_feature, metadata_name):
            builder.add_warning(
                header=f'Found invalid feature metadata "{metadata_name}" in config',
                subheader=(
                    f'The required metadata "{metadata_name}" is not a valid feature metadata.'
                    f" Please consider removing it from chalk.yml"
                ),
            )


def _validate_namespace_primary_key(namespace: str, features: List[UpsertFeatureGQL], builder: _LogBuilder):
    primary_features = list(f for f in features if f.scalarKind and f.scalarKind.primary)

    if len(primary_features) == 0:
        builder.add_error(
            header=f"Feature set '{namespace}' is missing a primary feature",
            subheader=f"Please add an 'int' or 'str' feature to '{namespace}', annotated with '= feature(primary=True)'",
        )
    elif len(primary_features) > 1:
        names = ", ".join([f.id.name for f in primary_features])
        builder.add_error(
            header=f"Feature set '{namespace}' has too many primary features",
            subheader=f"Found primary features: {names}. Composite primary keys are not supported. Please mark only a single feature as primary.",
        )


def validate_graph(request: UpsertGraphGQL) -> List[UpdateGraphError]:
    builder = _LogBuilder()

    # Validate the features
    _validate_feature_names_unique(request.features or [], builder)
    _validate_feature_metadata(request.features or [], builder, request.config)
    _validate_feature_metadata_config(request.features or [], builder, request.config)

    namespaces: Dict[str, List[UpsertFeatureGQL]] = defaultdict(list)

    for feature in request.features or []:
        namespaces[feature.id.namespace].append(feature)

    for namespace, features in namespaces.items():
        _validate_namespace_primary_key(namespace, features, builder)

    for feature in request.features or []:
        _validate_primary_key(feature, builder)
        _validate_max_staleness(feature, builder)
        _validate_joins(feature, builder)
        _validate_etl_to_online(feature, builder)
        _validate_feature_names(feature, builder)

    # Validate the resolvers
    fqn_to_feature = {f.id.fqn: f for f in request.features or []}
    _validate_resolver_names_unique(request.resolvers or [], builder)
    for resolver in request.resolvers or []:
        _validate_offline_resolver_has_one(resolver, builder)
        _validate_resolver_input(resolver, builder)
        _validate_resolver_output(resolver, builder)
        _validate_no_feature_times_as_input(
            fqn_to_feature=fqn_to_feature,
            resolver=resolver,
            builder=builder,
        )

    return builder.get_logs()



import re
from datetime import timedelta
from re import Pattern
from typing import Literal, Mapping, Optional, Tuple, Union

from typing_extensions import TypeAlias

Duration: TypeAlias = Union[str, timedelta, Literal["infinity"]]
"""
Duration is used to describe time periods in natural language.
To specify using natural language, write the count of the unit
you would like, followed by the representation of the unit.

Chalk supports the following units:

 | Signifier | Meaning       |
 | --------- | ------------- |
 | w         | Weeks         |
 | d         | Days          |
 | h         | Hours         |
 | m         | Minutes       |
 | s         | Seconds       |
 | ms        | Milliseconds  |

As well as the special keyword `"infinity"`

Examples


| Signifier   | Meaning                           |
| ----------- | --------------------------------- |
| "10h"       | 10 hours                          |
| "1w 2m"     | 1 week and 2 minutes              |
| "1h 10m 2s" | 1 hour, 10 minutes, and 2 seconds |
| "infinity"  | Unbounded time duration           |

Read more at https://docs.chalk.ai/docs/duration
"""

CronTab: TypeAlias = str
"""
A schedule defined using the Unix-cron
string format (`* * * * *`).
Values are given in the order below:


| Field        | Values |
| ------------ | ------ |
| Minute       | 0-59   |
| Hour         | 0-23   |
| Day of Month | 1-31   |
| Month        | 1-12   |
| Day of Week  | 0-6    |
"""

ScheduleOptions: TypeAlias = Optional[Union[CronTab, Duration, Literal[True]]]
"""The schedule on which to run a resolver.

One of:
- `CronTab`: A Unix-cron string, e.g. `"* * * * *"`.
- `Duration`: A Chalk Duration, e.g. `"2h 30m"`.
- `True`: Used internally.
"""


def timedelta_to_duration(t: timedelta) -> Duration:
    s = ""
    if t.days > 0:
        s += f"{t.days}d"
    if t.seconds > 0:
        if len(s) > 0:
            s += " "
        s += f"{t.seconds}s"
    return s


_meaning_to_signifier = dict(weeks="w", days="d", hours="h", minutes="m", seconds="s", milliseconds="ms")

_kwarg_to_regex: Mapping[str, Pattern] = {
    k: re.compile(rf"(([0-9]+?){v}(?![a-z]))") for k, v in _meaning_to_signifier.items()
}


def parse_chalk_duration(s: str) -> timedelta:
    def parse(regex: Pattern) -> Tuple[int, str]:
        matched = regex.search(s)
        if matched is None:
            return 0, ""

        return int(matched.groups()[1]), matched.groups()[0]

    parsed_values = {k: parse(unit) for k, unit in _kwarg_to_regex.items()}

    # Check for remaining unparsed input
    remainder = s
    for v in parsed_values.values():
        remainder = remainder.replace(v[1], "", 1)
    remainder = remainder.strip()

    if remainder != "":
        raise ValueError(f"Unparsed portion of duration: '{remainder}' for input: '{s}'")

    return timedelta(**{k: v[0] for k, v in parsed_values.items()})
